@@ -65,6 +65,7 @@
 #include <unistd.h>
 
 #include "libs/fvwmlib.h"
+#include "libs/FShape.h"
 #include "fvwm.h"
 #include "externs.h"
 #include "cursor.h"
@@ -76,9 +77,6 @@
 #include "events.h"
 #include "libs/Colorset.h"
 #include "fvwmsignal.h"
-#ifdef SHAPE
-#include <X11/extensions/shape.h>
-#endif /* SHAPE */
 #include "module_interface.h"
 #include "session.h"
 #include "borders.h"
@@ -117,12 +115,7 @@ static FvwmWindow *xcrossing_last_grab_window = NULL;
 
 
 STROKE_CODE(static int send_motion;)
-STROKE_CODE(static char sequence[MAX_SEQUENCE+1];)
-
-#ifdef SHAPE
-extern int ShapeEventBase;
-void HandleShapeNotify(void);
-#endif /* SHAPE */
+STROKE_CODE(static char sequence[STROKE_MAX_SEQUENCE + 1];)
 
 Window PressedW = None;
 
@@ -740,7 +733,7 @@ ICON_DBG((stderr,"hpn: applying new icon '%s'\n", Tmp_win->name));
 	  Tmp_win->hints.win_gravity, &Tmp_win->normal_g, wdiff, hdiff);
       }
       gravity_constrain_size(
-	Tmp_win->hints.win_gravity, Tmp_win, &Tmp_win->normal_g);
+	Tmp_win->hints.win_gravity, Tmp_win, &Tmp_win->normal_g, 0);
       if (!IS_MAXIMIZED(Tmp_win))
       {
 	rectangle new_g;
@@ -769,7 +762,8 @@ ICON_DBG((stderr,"hpn: applying new icon '%s'\n", Tmp_win->name));
 	w = Tmp_win->max_g.width;
 	h = Tmp_win->max_g.height;
 	gravity_constrain_size(
-	  Tmp_win->hints.win_gravity, Tmp_win, &Tmp_win->max_g);
+	  Tmp_win->hints.win_gravity, Tmp_win, &Tmp_win->max_g,
+	  CS_UPDATE_MAX_DEFECT);
 	if (w != Tmp_win->max_g.width ||
 	    h != Tmp_win->max_g.height)
 	{
@@ -1448,6 +1442,12 @@ void HandleReparentNotify(void)
 {
   if (!Tmp_win)
     return;
+  if (Event.xreparent.parent == Scr.Root)
+  {
+    /* Ignore reparenting to the root window.  In some cases these events are
+     * selected although the window is no longer managed. */
+    return;
+  }
   if (Event.xreparent.parent != Tmp_win->frame)
   {
     /* window was reparented by someone else, destroy the frame */
@@ -1482,6 +1482,10 @@ void HandleButtonPress(void)
 
   GrabEm(CRS_NONE, GRAB_PASSIVE);
   if (!Tmp_win &&
+      (Event.xbutton.window != Scr.PanFrameTop.win &&
+       Event.xbutton.window != Scr.PanFrameBottom.win &&
+       Event.xbutton.window != Scr.PanFrameLeft.win &&
+       Event.xbutton.window != Scr.PanFrameRight.win) &&
       (Event.xbutton.window != Scr.Root || Event.xbutton.subwindow != None))
   {
     /* event in unmanaged window or subwindow of a client */
@@ -1566,14 +1570,13 @@ void HandleButtonPress(void)
 	  get_focus_window() != Tmp_win)
       {
 	XAllowEvents(dpy,ReplayPointer,CurrentTime);
+	/*	return;*/
       }
       else /* don't pass click to just focused window */
       {
 	XAllowEvents(dpy,AsyncPointer,CurrentTime);
       }
-      XSync(dpy,0);
       UngrabEm(GRAB_PASSIVE);
-      return;
     }
     if (!IS_ICONIFIED(Tmp_win))
     {
@@ -1809,7 +1812,7 @@ void HandleEnterNotify(void)
     if (is_initial_ungrab_pending)
     {
       is_initial_ungrab_pending = False;
-      xcrossing_last_grab_window = None;
+      xcrossing_last_grab_window = NULL;
     }
     else if (Tmp_win && Tmp_win == xcrossing_last_grab_window)
     {
@@ -1833,7 +1836,9 @@ void HandleEnterNotify(void)
       StashEventTime(&d);
       if((d.xcrossing.mode==NotifyNormal)&&
 	 (d.xcrossing.detail!=NotifyInferior))
+      {
 	return;
+      }
     }
   }
 
@@ -1898,7 +1903,7 @@ void HandleEnterNotify(void)
   }
 
   sf = get_focus_window();
-  if (sf && HAS_MOUSE_FOCUS(sf))
+  if (sf && Tmp_win != sf && HAS_MOUSE_FOCUS(sf))
   {
     DeleteFocus(1);
   }
@@ -1958,7 +1963,9 @@ void HandleLeaveNotify(void)
   /* Ignore LeaveNotify events while a window is resized or moved as a wire
    * frame; otherwise the window list may be screwed up. */
   if (Scr.flags.is_wire_frame_displayed)
+  {
     return;
+  }
   if (Event.xcrossing.mode != NotifyNormal)
   {
     /* Ignore events generated by grabbing or ungrabbing the pointer.  However,
@@ -2191,15 +2198,20 @@ void HandleConfigureRequest(void)
       return;
   }
 
-#ifdef SHAPE
-  if (ShapesSupported)
+  if (FHaveShapeExtension && FShapesSupported)
   {
-    int xws, yws, xbs, ybs;
-    unsigned wws, hws, wbs, hbs;
-    int boundingShaped, clipShaped;
+    int i;
+    unsigned int u;
+    Bool b;
+    int boundingShaped;
 
-    if (XShapeQueryExtents(dpy, Tmp_win->w,&boundingShaped, &xws, &yws, &wws,
-			   &hws,&clipShaped, &xbs, &ybs, &wbs, &hbs))
+    /* suppress compiler warnings w/o shape extension */
+    i = 0;
+    u = 0;
+    b = False;
+
+    if (FShapeQueryExtents(
+	  dpy, Tmp_win->w, &boundingShaped, &i, &i, &u, &u, &b, &i, &i, &u, &u))
     {
       Tmp_win->wShaped = boundingShaped;
     }
@@ -2208,7 +2220,6 @@ void HandleConfigureRequest(void)
       Tmp_win->wShaped = 0;
     }
   }
-#endif /* SHAPE */
 
   if (cre->window == Tmp_win->w)
   {
@@ -2286,8 +2297,7 @@ fprintf(stderr, "cre: %d(%d) %d(%d) %d(%d)x%d(%d) w 0x%08x '%s'\n",
     constr_w = oldnew_w;
     constr_h = oldnew_h;
     constrain_size(
-      Tmp_win, (unsigned int *)&constr_w, (unsigned int *)&constr_h, 0, 0,
-      False);
+      Tmp_win, (unsigned int *)&constr_w, (unsigned int *)&constr_h, 0, 0, 0);
     dw += (constr_w - oldnew_w);
     dh += (constr_h - oldnew_h);
     if ((cre->value_mask & CWX) && dw)
@@ -2496,24 +2506,22 @@ void SendConfigureNotify(
  *      HandleShapeNotify - shape notification event handler
  *
  ***********************************************************************/
-#ifdef SHAPE
 void HandleShapeNotify (void)
 {
   DBUG("HandleShapeNotify","Routine Entered");
 
-  if (ShapesSupported)
+  if (FShapesSupported)
   {
-    XShapeEvent *sev = (XShapeEvent *) &Event;
+    FShapeEvent *sev = (FShapeEvent *) &Event;
 
     if (!Tmp_win)
       return;
-    if (sev->kind != ShapeBounding)
+    if (sev->kind != FShapeBounding)
       return;
     Tmp_win->wShaped = sev->shaped;
-    SetShape(Tmp_win,Tmp_win->frame_g.width);
+    SetShape(Tmp_win, Tmp_win->frame_g.width);
   }
 }
-#endif  /* SHAPE*/
 
 /***********************************************************************
  *
@@ -2732,10 +2740,8 @@ void InitEventHandlerJumpTable(void)
   EventHandlerJumpTable[KeyPress] =         HandleKeyPress;
   EventHandlerJumpTable[VisibilityNotify] = HandleVisibilityNotify;
   EventHandlerJumpTable[ColormapNotify] =   HandleColormapNotify;
-#ifdef SHAPE
-  if (ShapesSupported)
-    EventHandlerJumpTable[ShapeEventBase+ShapeNotify] = HandleShapeNotify;
-#endif /* SHAPE */
+  if (FShapesSupported)
+    EventHandlerJumpTable[FShapeEventBase+FShapeNotify] = HandleShapeNotify;
   EventHandlerJumpTable[SelectionClear]   = HandleSelectionClear;
   EventHandlerJumpTable[SelectionRequest] = HandleSelectionRequest;
   EventHandlerJumpTable[ReparentNotify] = HandleReparentNotify;
@@ -2872,8 +2878,10 @@ int GetContext(FvwmWindow *t, XEvent *e, Window *w)
 	&(e->xkey.y), &(e->xkey.subwindow));
     }
   }
-  if(!t)
+  if (!t)
+  {
     return C_ROOT;
+  }
 
   if (e->type == KeyPress && e->xkey.window == t->frame &&
       e->xkey.subwindow == t->decor_w)
@@ -2889,7 +2897,9 @@ int GetContext(FvwmWindow *t, XEvent *e, Window *w)
   *w= e->xany.window;
 
   if (*w == Scr.NoFocusWin)
+  {
     return C_ROOT;
+  }
   if (e->type == KeyPress && e->xkey.window == t->frame &&
       e->xkey.subwindow == t->decor_w)
   {
@@ -2904,7 +2914,9 @@ int GetContext(FvwmWindow *t, XEvent *e, Window *w)
   *w= e->xany.window;
 
   if (*w == Scr.NoFocusWin)
+  {
     return C_ROOT;
+  }
   if (e->xkey.subwindow != None &&
       (e->xkey.window == t->decor_w || e->xkey.window == t->Parent))
   {
@@ -2915,52 +2927,54 @@ int GetContext(FvwmWindow *t, XEvent *e, Window *w)
     e->xkey.subwindow = None;
   }
   if (*w == Scr.Root)
+  {
     return C_ROOT;
+  }
   if (t)
+  {
+    if (*w == t->title_w)
+      Context = C_TITLE;
+    else if (*w == t->w || *w == t->Parent || *w == t->frame)
+      Context = C_WINDOW;
+    else if (*w == t->icon_w || *w == t->icon_pixmap_w)
+      Context = C_ICON;
+    else if (*w == t->decor_w)
+      Context = C_SIDEBAR;
+    else
     {
-      if (*w == t->title_w)
-	Context = C_TITLE;
-      else if (*w == t->w || *w == t->Parent || *w == t->frame)
-	Context = C_WINDOW;
-      else if (*w == t->icon_w || *w == t->icon_pixmap_w)
-	Context = C_ICON;
-      else if (*w == t->decor_w)
-	Context = C_SIDEBAR;
+      for(i=0;i<4;i++)
+      {
+	if(*w == t->corners[i])
+	{
+	  Context = C_FRAME;
+	  break;
+	}
+	if(*w == t->sides[i])
+	{
+	  Context = C_SIDEBAR;
+	  break;
+	}
+      }
+      if (i < 4)
+	Button = i;
       else
       {
-	for(i=0;i<4;i++)
+	for (i = 0; i < NUMBER_OF_BUTTONS; i++)
+	{
+	  if (*w == t->button_w[i])
 	  {
-	    if(*w == t->corners[i])
-	      {
-		Context = C_FRAME;
-		break;
-	      }
-	    if(*w == t->sides[i])
-	      {
-		Context = C_SIDEBAR;
-		break;
-	      }
-	  }
-	if (i < 4)
-	  Button = i;
-	else
-	  {
-	    for (i = 0; i < NUMBER_OF_BUTTONS; i++)
+	    if ((!(i & 1) && i / 2 < Scr.nr_left_buttons) ||
+		( (i & 1) && i / 2 < Scr.nr_right_buttons))
 	    {
-	      if (*w == t->button_w[i])
-	      {
-		if ((!(i & 1) && i / 2 < Scr.nr_left_buttons) ||
-		    ( (i & 1) && i / 2 < Scr.nr_right_buttons))
-		{
-		  Context = (1 << i) * C_L1;
-		  Button = i;
-		  break;
-		}
-	      }
+	      Context = (1 << i) * C_L1;
+	      Button = i;
+	      break;
 	    }
 	  }
-      } /* else */
-    } /* if (t) */
+	}
+      }
+    } /* else */
+  } /* if (t) */
   return Context;
 }
 
