@@ -38,6 +38,8 @@
 #endif /* XPM */
 
 #include "libs/fvwmlib.h"
+#include "libs/FScreen.h"
+#include "libs/FShape.h"
 #include <stdio.h>
 #include "fvwm.h"
 #include "externs.h"
@@ -60,15 +62,11 @@
 #include "gnome.h"
 #include "geometry.h"
 
-#ifdef SHAPE
-#include <X11/extensions/shape.h>
-#endif /* SHAPE */
-
 static int do_all_iconboxes(FvwmWindow *t, icon_boxes **icon_boxes_ptr);
 static void GetBitmapFile(FvwmWindow *tmp_win);
+#ifdef XPM
 static void GetXPMFile(FvwmWindow *tmp_win);
-
-
+#endif
 
 /****************************************************************************
  *
@@ -82,7 +80,16 @@ void CreateIconWindow(FvwmWindow *tmp_win, int def_x, int def_y)
   /* attributes for create windows */
   XSetWindowAttributes attributes;
   XWindowChanges xwc;
+  Window old_icon_pixmap_w;
+  Window old_icon_w;
+  Bool is_old_icon_shaped = IS_ICON_SHAPED(tmp_win);
+  char icon_order[4];
+  int i;
 
+  old_icon_w = tmp_win->icon_w;
+  old_icon_pixmap_w = (IS_ICON_OURS(tmp_win)) ? tmp_win->icon_pixmap_w : None;
+  if (!IS_ICON_OURS(tmp_win) && tmp_win->icon_pixmap_w)
+    XUnmapWindow(dpy,tmp_win->icon_pixmap_w);
   SET_ICON_OURS(tmp_win, 1);
   SET_PIXMAP_OURS(tmp_win, 0);
   SET_ICON_SHAPED(tmp_win, 0);
@@ -94,29 +101,103 @@ void CreateIconWindow(FvwmWindow *tmp_win, int def_x, int def_y)
     return;
 
   /* First, see if it was specified in the .fvwmrc */
+  if (ICON_OVERRIDE_MODE(tmp_win) == ICON_OVERRIDE)
+  {
+    /* try fvwm provided icons before application provided icons */
+    icon_order[0] = 0;
+    icon_order[1] = 1;
+    icon_order[2] = 2;
+    icon_order[3] = 3;
+ICON_DBG((stderr,"ciw: hint order: xpm bmp iwh iph '%s'\n", tmp_win->name));
+  }
+  else if (ICON_OVERRIDE_MODE(tmp_win) == NO_ACTIVE_ICON_OVERRIDE)
+  {
+    if (tmp_win->wmhints && (tmp_win->wmhints->flags & IconPixmapHint) &&
+	WAS_ICON_HINT_PROVIDED(tmp_win) == ICON_HINT_MULTIPLE)
+    {
+      /* use application provided icon window or pixmap first, then fvwm
+       * provided icons. */
+      icon_order[0] = 2;
+      icon_order[1] = 3;
+      icon_order[2] = 0;
+      icon_order[3] = 1;
+ICON_DBG((stderr,"ciw: hint order: iwh iph xpm bmp '%s'\n", tmp_win->name));
+    }
+    else if (Scr.DefaultIcon && tmp_win->icon_bitmap_file == Scr.DefaultIcon)
+    {
+      /* use application provided icon window/pixmap first, then fvwm provided
+       * default icon */
+      icon_order[0] = 2;
+      icon_order[1] = 3;
+      icon_order[2] = 0;
+      icon_order[3] = 1;
+ICON_DBG((stderr,"ciw: hint order: iwh iph xpm bmp '%s'\n", tmp_win->name));
+    }
+    else
+    {
+      /* use application provided icon window first, then fvwm provided icons
+       * and then application provided icon pixmap */
+      icon_order[0] = 2;
+      icon_order[1] = 0;
+      icon_order[2] = 1;
+      icon_order[3] = 3;
+ICON_DBG((stderr,"ciw: hint order: iwh xpm bmp iph '%s'\n", tmp_win->name));
+    }
+  }
+  else
+  {
+    /* use application provided icon rather than fvwm provided icon */
+    icon_order[0] = 2;
+    icon_order[1] = 3;
+    icon_order[2] = 0;
+    icon_order[3] = 1;
+ICON_DBG((stderr,"ciw: hint order: iwh iph bmp xpm '%s'\n", tmp_win->name));
+  }
   tmp_win->icon_p_height = 0;
   tmp_win->icon_p_width = 0;
-
-  /* First, check for a monochrome bitmap */
-  if (tmp_win->icon_bitmap_file != NULL)
-    GetBitmapFile(tmp_win);
-
+  for (i = 0; i < 4 && !tmp_win->icon_p_height && !tmp_win->icon_p_width; i++)
+  {
+    switch (icon_order[i])
+    {
+    case 0:
 #ifdef XPM
-  /* Next, check for a color pixmap */
-  if ((tmp_win->icon_bitmap_file != NULL) && (tmp_win->icon_p_height == 0)
-      && (tmp_win->icon_p_width == 0))
-    GetXPMFile(tmp_win);
+      /* Next, check for a color pixmap */
+      if (tmp_win->icon_bitmap_file)
+      {
+	GetXPMFile(tmp_win);
+      }
+ICON_DBG((stderr,"ciw: xpm%s used '%s'\n", (tmp_win->icon_p_height)?"":" not",tmp_win->name));
 #endif /* XPM */
-
-  /* Next, See if the app supplies its own icon window */
-  if ((tmp_win->icon_p_height == 0) && (tmp_win->icon_p_width == 0)
-      && (tmp_win->wmhints) && (tmp_win->wmhints->flags & IconWindowHint))
-    GetIconWindow(tmp_win);
-
-  /* Finally, try to get icon bitmap from the application */
-  if ((tmp_win->icon_p_height == 0) && (tmp_win->icon_p_width == 0)
-      && (tmp_win->wmhints) && (tmp_win->wmhints->flags & IconPixmapHint))
-    GetIconBitmap(tmp_win);
+      break;
+    case 1:
+      /* First, check for a monochrome bitmap */
+      if (tmp_win->icon_bitmap_file)
+      {
+	GetBitmapFile(tmp_win);
+      }
+ICON_DBG((stderr,"ciw: bmp%s used '%s'\n", (tmp_win->icon_p_height)?"":" not",tmp_win->name));
+      break;
+    case 2:
+      /* Next, See if the app supplies its own icon window */
+      if (tmp_win->wmhints && (tmp_win->wmhints->flags & IconWindowHint))
+      {
+	GetIconWindow(tmp_win);
+      }
+ICON_DBG((stderr,"ciw: iwh%s used '%s'\n", (tmp_win->icon_p_height)?"":" not",tmp_win->name));
+      break;
+    case 3:
+      /* Finally, try to get icon bitmap from the application */
+      if (tmp_win->wmhints && (tmp_win->wmhints->flags & IconPixmapHint))
+      {
+	GetIconBitmap(tmp_win);
+      }
+ICON_DBG((stderr,"ciw: iph%s used '%s'\n", (tmp_win->icon_p_height)?"":" not",tmp_win->name));
+      break;
+    default:
+      /* can't happen */
+      break;
+    }
+  }
 
   /* figure out the icon window size */
   if (!HAS_NO_ICON_TITLE(tmp_win))
@@ -156,17 +237,31 @@ void CreateIconWindow(FvwmWindow *tmp_win, int def_x, int def_y)
   attributes.background_pixel = Scr.StdBack;
   attributes.cursor = Scr.FvwmCursors[CRS_DEFAULT];
   attributes.border_pixel = 0;
-  attributes.event_mask = (ButtonPressMask | ButtonReleaseMask
-			   | VisibilityChangeMask | ExposureMask | KeyPressMask
-			   | EnterWindowMask | LeaveWindowMask
-			   | FocusChangeMask );
-  if (!HAS_NO_ICON_TITLE(tmp_win))
+  attributes.event_mask = XEVMASK_ICONW;
+  if (HAS_NO_ICON_TITLE(tmp_win))
   {
-    tmp_win->icon_w =
-      XCreateWindow(
-	dpy, Scr.Root, def_x, def_y + tmp_win->icon_p_height,
-	tmp_win->icon_g.width, tmp_win->icon_g.height, 0, Pdepth,
-	InputOutput, Pvisual, valuemask, &attributes);
+    if (tmp_win->icon_w)
+    {
+      XDestroyWindow(dpy, tmp_win->icon_w);
+      XDeleteContext(dpy, tmp_win->icon_w, FvwmContext);
+    }
+  }
+  else
+  {
+    if (!tmp_win->icon_w)
+    {
+      tmp_win->icon_w =
+	XCreateWindow(
+	  dpy, Scr.Root, def_x, def_y + tmp_win->icon_p_height,
+	  tmp_win->icon_g.width, tmp_win->icon_g.height, 0, Pdepth,
+	  InputOutput, Pvisual, valuemask, &attributes);
+    }
+    else
+    {
+      XMoveResizeWindow(
+	dpy, tmp_win->icon_w, def_x, def_y + tmp_win->icon_p_height,
+	  tmp_win->icon_g.width, tmp_win->icon_g.height);
+    }
   }
   if (Scr.DefaultColorset >= 0)
     SetWindowBackground(dpy, tmp_win->icon_w, tmp_win->icon_g.width,
@@ -174,17 +269,27 @@ void CreateIconWindow(FvwmWindow *tmp_win, int def_x, int def_y)
 			Pdepth, Scr.StdGC, False);
 
   /* create a window to hold the picture */
-  if((IS_ICON_OURS(tmp_win)) && (tmp_win->icon_p_width > 0)
-     && (tmp_win->icon_p_height > 0))
+  if (IS_ICON_OURS(tmp_win) && tmp_win->icon_p_width > 0 &&
+      tmp_win->icon_p_height > 0)
   {
     /* use fvwm's visuals in these cases */
     if (Pdefault || (tmp_win->iconDepth == 1) || IS_PIXMAP_OURS(tmp_win))
     {
-      tmp_win->icon_pixmap_w =
-	XCreateWindow(
-	  dpy, Scr.Root, def_x, def_y, tmp_win->icon_p_width,
-	  tmp_win->icon_p_height, 0, Pdepth, InputOutput, Pvisual, valuemask,
-	  &attributes);
+      if (!old_icon_pixmap_w)
+      {
+	tmp_win->icon_pixmap_w =
+	  XCreateWindow(
+	    dpy, Scr.Root, def_x, def_y, tmp_win->icon_p_width,
+	    tmp_win->icon_p_height, 0, Pdepth, InputOutput, Pvisual, valuemask,
+	    &attributes);
+      }
+      else
+      {
+	tmp_win->icon_pixmap_w = old_icon_pixmap_w;
+	XMoveResizeWindow(
+	  dpy, tmp_win->icon_pixmap_w, def_x, def_y, tmp_win->icon_p_width,
+	  tmp_win->icon_p_height);
+      }
       if (Scr.DefaultColorset >= 0)
 	SetWindowBackground(dpy, tmp_win->icon_w, tmp_win->icon_p_width,
 			    tmp_win->icon_p_height,
@@ -212,46 +317,64 @@ void CreateIconWindow(FvwmWindow *tmp_win, int def_x, int def_y)
   else
   {
     /* client supplied icon window: select events on it */
-    attributes.event_mask = ButtonPressMask | ButtonReleaseMask | KeyPressMask
-      | VisibilityChangeMask | FocusChangeMask
-      | EnterWindowMask | LeaveWindowMask;
+    attributes.event_mask = XEVMASK_ICONPW;
     valuemask = CWEventMask;
     XChangeWindowAttributes(dpy, tmp_win->icon_pixmap_w, valuemask,&attributes);
   }
-
-
-#ifdef SHAPE
-  if (ShapesSupported && IS_ICON_SHAPED(tmp_win))
+  if (old_icon_pixmap_w && old_icon_pixmap_w != tmp_win->icon_pixmap_w)
   {
-    /* when fvwm is using the non-default visual client supplied icon pixmaps
-     * are drawn in a window with no relief */
-    int off = (Pdefault || (tmp_win->iconDepth == 1) || IS_PIXMAP_OURS(tmp_win))
-      ? 2 : 0;
-    XShapeCombineMask(dpy, tmp_win->icon_pixmap_w, ShapeBounding, off, off,
-		      tmp_win->icon_maskPixmap, ShapeSet);
+    /* destroy the old window */
+    XDestroyWindow(dpy, old_icon_pixmap_w);
+    XDeleteContext(dpy, old_icon_pixmap_w, FvwmContext);
+    is_old_icon_shaped = False;
   }
-#endif
 
-  if(tmp_win->icon_w != None)
+  if (FShapesSupported)
+  {
+    if (IS_ICON_SHAPED(tmp_win))
+    {
+      /* when fvwm is using the non-default visual client supplied icon pixmaps
+       * are drawn in a window with no relief */
+      int off;
+      off = (Pdefault || (tmp_win->iconDepth == 1) || IS_PIXMAP_OURS(tmp_win)) ?
+	2 : 0;
+      FShapeCombineMask(
+	dpy, tmp_win->icon_pixmap_w, FShapeBounding, off, off,
+	tmp_win->icon_maskPixmap, FShapeSet);
+    }
+    else if (is_old_icon_shaped && tmp_win->icon_pixmap_w == old_icon_pixmap_w)
+    {
+      /* remove the shape */
+      XRectangle r;
+
+      r.x = 0;
+      r.y = 0;
+      r.width = tmp_win->icon_p_width;
+      r.height = tmp_win->icon_p_height;
+      FShapeCombineRectangles(
+	dpy, tmp_win->icon_pixmap_w, FShapeBounding, 0, 0, &r, 1, FShapeSet, 0);
+    }
+  }
+
+  if (tmp_win->icon_w != None && tmp_win->icon_w != old_icon_w)
   {
     XSaveContext(dpy, tmp_win->icon_w, FvwmContext, (caddr_t)tmp_win);
     XDefineCursor(dpy, tmp_win->icon_w, Scr.FvwmCursors[CRS_DEFAULT]);
     GrabAllWindowKeysAndButtons(dpy, tmp_win->icon_w, Scr.AllBindings,
 				C_ICON, GetUnusedModifiers(),
 				Scr.FvwmCursors[CRS_DEFAULT], True);
-
     xwc.sibling = tmp_win->frame;
     xwc.stack_mode = Below;
     XConfigureWindow(dpy, tmp_win->icon_w, CWSibling|CWStackMode, &xwc);
   }
-  if(tmp_win->icon_pixmap_w != None)
+  if (tmp_win->icon_pixmap_w != None &&
+      tmp_win->icon_pixmap_w != old_icon_pixmap_w)
   {
     XSaveContext(dpy, tmp_win->icon_pixmap_w, FvwmContext, (caddr_t)tmp_win);
     XDefineCursor(dpy, tmp_win->icon_pixmap_w, Scr.FvwmCursors[CRS_DEFAULT]);
     GrabAllWindowKeysAndButtons(dpy, tmp_win->icon_pixmap_w, Scr.AllBindings,
 				C_ICON, GetUnusedModifiers(),
 				Scr.FvwmCursors[CRS_DEFAULT], True);
-
     xwc.sibling = tmp_win->frame;
     xwc.stack_mode = Below;
     XConfigureWindow(dpy,tmp_win->icon_pixmap_w,CWSibling|CWStackMode,&xwc);
@@ -312,6 +435,11 @@ void DrawIconWindow(FvwmWindow *tmp_win)
   {
     if (IS_ICON_ENTERED(tmp_win))
     {
+      int sx;
+      int sy;
+      int sw;
+      int sh;
+
       /* resize the icon name window */
       tmp_win->icon_g.width = tmp_win->icon_t_width + 8;
       if (IS_STICKY(tmp_win) || IS_ICON_STICKY(tmp_win))
@@ -320,15 +448,16 @@ void DrawIconWindow(FvwmWindow *tmp_win)
 	tmp_win->icon_g.width = tmp_win->icon_p_width;
       tmp_win->icon_xl_loc = tmp_win->icon_g.x -
 	(tmp_win->icon_g.width - tmp_win->icon_p_width)/2;
+      FScreenGetScrRect(NULL, FSCREEN_CURRENT, &sx, &sy, &sw, &sh);
       /* start keep label on screen. dje 8/7/97 */
-      if (tmp_win->icon_xl_loc < 0) {  /* if new loc neg (off left edge) */
-	tmp_win->icon_xl_loc = 0;      /* move to edge */
+      if (tmp_win->icon_xl_loc < sx) {  /* if new loc neg (off left edge) */
+	tmp_win->icon_xl_loc = sx;      /* move to edge */
       } else {                         /* if not on left edge */
 	/* if (new loc + width) > screen width (off edge on right) */
 	if ((tmp_win->icon_xl_loc + tmp_win->icon_g.width) >
-	    Scr.MyDisplayWidth) {      /* off right */
+	    sx + sw) {      /* off right */
 	  /* position up against right edge */
-	  tmp_win->icon_xl_loc = Scr.MyDisplayWidth-tmp_win->icon_g.width;
+	  tmp_win->icon_xl_loc = sx + sw - tmp_win->icon_g.width;
 	}
 	/* end keep label on screen. dje 8/7/97 */
       }
@@ -475,6 +604,9 @@ void DrawIconWindow(FvwmWindow *tmp_win)
       XConfigureWindow(dpy, tmp_win->icon_pixmap_w, mask, &xwc);
     }
   }
+  /* wait for pending EnterNotify/LeaveNotify events to suppress race condition
+   * w/ expanding/collapsing icon titles */
+  XSync(dpy, 0);
 }
 
 
@@ -500,7 +632,9 @@ void RedoIconName(FvwmWindow *tmp_win)
                  strlen(tmp_win->icon_name));
   /* clear the icon window, and trigger a re-draw via an expose event */
   if (IS_ICONIFIED(tmp_win))
+  {
     DrawIconWindow(tmp_win);
+  }
   if (IS_ICONIFIED(tmp_win))
     XClearArea(dpy, tmp_win->icon_w, 0, 0, 0, 0, True);
   return;
@@ -522,6 +656,8 @@ void AutoPlaceIcon(FvwmWindow *t)
   int width,height;
   FvwmWindow *test_window;
   Bool loc_ok;
+  Bool loc_ok_wrong_screen;
+  Bool loc_ok_wrong_screen2;
   int real_x=10, real_y=10;
   int new_x, new_y;
   Bool do_move_pixmap = False;
@@ -530,7 +666,7 @@ void AutoPlaceIcon(FvwmWindow *t)
   /* New! Put icon in same page as the center of the window */
   /* Not a good idea for StickyIcons. Neither for icons of windows that are
    * visible on the current page. */
-  if(IS_ICON_STICKY(t)||IS_STICKY(t))
+  if (IS_ICON_STICKY(t) || IS_STICKY(t))
   {
     base_x = 0;
     base_y = 0;
@@ -569,7 +705,7 @@ void AutoPlaceIcon(FvwmWindow *t)
   }
   if(IS_ICON_MOVED(t))
   {
-    /* just make sure the icon is on this screen */
+    /* just make sure the icon is on this page */
     t->icon_g.x = t->icon_g.x % Scr.MyDisplayWidth + base_x;
     t->icon_g.y = t->icon_g.y % Scr.MyDisplayHeight + base_y;
     if(t->icon_g.x < 0)
@@ -591,6 +727,38 @@ void AutoPlaceIcon(FvwmWindow *t)
      Look thru chain of icon boxes assigned to window.
      Add logic for grids and fill direction.
   */
+  else if (DO_IGNORE_ICON_BOXES(t))
+  {
+    int sx;
+    int sy;
+    int sw;
+    int sh;
+    fscreen_scr_arg fscr;
+
+    t->icon_g.x = t->frame_g.x;
+    if (HAS_BOTTOM_TITLE(t))
+    {
+      t->icon_g.y =
+	t->frame_g.y + t->frame_g.height - t->icon_g.height - t->icon_p_height;
+    }
+    else
+    {
+      t->icon_g.y = t->frame_g.y;
+    }
+    fscr.xypos.x = t->icon_g.x + t->icon_g.width / 2;
+    fscr.xypos.y = t->icon_g.y + t->icon_g.height / 2;
+    FScreenGetScrRect(&fscr, FSCREEN_XYPOS, &sx, &sy, &sw, &sh);
+    if (t->icon_g.x < sx)
+      t->icon_g.x = sx;
+    else if (t->icon_g.x + t->icon_g.width > sx + sw)
+      t->icon_g.x = sx + sw - t->icon_g.width;
+    if (t->icon_g.y < sy)
+      t->icon_g.y = sy;
+    else if (t->icon_g.y + t->icon_g.height + t->icon_p_height > sy + sh)
+      t->icon_g.y = sy + sh - t->icon_g.height - t->icon_p_height;
+    do_move_pixmap = True;
+    do_move_title = True;
+  }
   else
   {
     /* A place to hold inner and outer loop variables. */
@@ -604,10 +772,13 @@ void AutoPlaceIcon(FvwmWindow *t)
       int icon_dimension;               /* height or width */
       int nom_dimension;                /* nonminal height or width */
       int screen_dimension;             /* screen height or width */
+      int screen_offset;                /* screen offset */
     } dimension;
     dimension dim[3];                   /* space for work, 1st, 2nd dimen */
     icon_boxes *icon_boxes_ptr;         /* current icon box */
     int i;                              /* index for inner/outer loop data */
+    fscreen_scr_arg fscr;
+    rectangle ref;
 
     /* Hopefully this makes the following more readable. */
 #define ICONBOX_LFT icon_boxes_ptr->IconBox[0]
@@ -618,12 +789,16 @@ void AutoPlaceIcon(FvwmWindow *t)
 #define RGT_FILL icon_boxes_ptr->IconFlags & ICONFILLRGT
 #define HRZ_FILL icon_boxes_ptr->IconFlags & ICONFILLHRZ
 
+    /* needed later */
+    fscr.xypos.x = t->frame_g.x + (t->frame_g.width / 2) - base_x;
+    fscr.xypos.y = t->frame_g.y + (t->frame_g.height / 2) - base_y;
     /* unnecessary copy of width */
     width = t->icon_p_width;
     /* total height */
     height = t->icon_g.height + t->icon_p_height;
     /* no slot found yet */
     loc_ok = False;
+    loc_ok_wrong_screen = False;
 
     /* check all boxes in order */
     icon_boxes_ptr = NULL;              /* init */
@@ -634,17 +809,33 @@ void AutoPlaceIcon(FvwmWindow *t)
 	/* leave for loop */
         break;
       }
+      /* get the screen dimensions for the icon box */
+      if (icon_boxes_ptr->IconScreen == FSCREEN_CURRENT)
+	fscr.mouse_ev = NULL;
+      FScreenGetScrRect(
+	&fscr,
+        icon_boxes_ptr->IconScreen,
+	&ref.x, &ref.y, &ref.width, &ref.height);
+      dim[1].screen_offset = ref.y;
+      dim[1].screen_dimension = ref.height;
+      dim[2].screen_offset = ref.x;
+      dim[2].screen_dimension = ref.width;
       /* y amount */
       dim[1].step = icon_boxes_ptr->IconGrid[1];
       /* init start from */
-      dim[1].start_at = ICONBOX_TOP;
+      dim[1].start_at = ICONBOX_TOP + dim[1].screen_offset;
+      if (icon_boxes_ptr->IconSign[1] == '-') {
+        dim[1].start_at += dim[1].screen_dimension;
+      }
       /* init end at */
-      dim[1].end_at = ICONBOX_BOT;
+      dim[1].end_at = ICONBOX_BOT + dim[1].screen_offset;
+      if (icon_boxes_ptr->IconSign[3] == '-') {
+        dim[1].end_at += dim[1].screen_dimension;
+      }
       /* save base */
       dim[1].base = base_y;
       /* save dimension */
       dim[1].icon_dimension = height;
-      dim[1].screen_dimension = Scr.MyDisplayHeight;
       if (BOT_FILL)
       {
 	/* fill from bottom */
@@ -655,14 +846,19 @@ void AutoPlaceIcon(FvwmWindow *t)
       /* x amount */
       dim[2].step = icon_boxes_ptr->IconGrid[0];
       /* init start from */
-      dim[2].start_at = ICONBOX_LFT;
+      dim[2].start_at = ICONBOX_LFT + dim[2].screen_offset;
+      if (icon_boxes_ptr->IconSign[0] == '-') {
+        dim[2].start_at += dim[2].screen_dimension;
+      }
       /* init end at */
-      dim[2].end_at = ICONBOX_RGT;
+      dim[2].end_at = ICONBOX_RGT + dim[2].screen_offset;
+      if (icon_boxes_ptr->IconSign[2] == '-') {
+        dim[2].end_at += dim[2].screen_dimension;
+      }
       /* save base */
       dim[2].base   = base_x;
       /* save dimension */
       dim[2].icon_dimension = width;
-      dim[2].screen_dimension = Scr.MyDisplayWidth;
       if (RGT_FILL)
       {
 	/* fill from right */
@@ -707,24 +903,26 @@ void AutoPlaceIcon(FvwmWindow *t)
       } /* end horizontal dimension first */
       /* save for reseting inner loop */
       dim[0].start_at = dim[2].start_at;
+      loc_ok_wrong_screen2 = False;
       while((dim[1].step < 0            /* filling reversed */
              ? (dim[1].start_at + dim[1].icon_dimension - dim[1].nom_dimension
                 > dim[1].end_at)        /* check back edge */
              : (dim[1].start_at + dim[1].nom_dimension
                 < dim[1].end_at))       /* check front edge */
-            && (!loc_ok)) {             /* nothing found yet */
+            && (!loc_ok)
+	    && (!loc_ok_wrong_screen2)) {  /* nothing found yet */
         dim[1].real_start = dim[1].start_at; /* init */
         if (dim[1].start_at + dim[1].icon_dimension >
-            dim[1].screen_dimension - 2 + dim[1].base)
+	    dim[1].screen_offset + dim[1].screen_dimension - 2 + dim[1].base)
 	{
 	  /* off screen, move on screen */
-          dim[1].real_start = dim[1].screen_dimension
+          dim[1].real_start = dim[1].screen_offset + dim[1].screen_dimension
             - dim[1].icon_dimension + dim[1].base;
         } /* end off screen */
-        if (dim[1].start_at < dim[1].base)
+        if (dim[1].start_at < dim[1].screen_offset + dim[1].base)
 	{
 	  /* if off other edge, move on screen */
-          dim[1].real_start = dim[1].base;
+          dim[1].real_start = dim[1].screen_offset + dim[1].base;
         } /* end off other edge */
 	/* reset inner loop */
         dim[2].start_at = dim[0].start_at;
@@ -733,19 +931,20 @@ void AutoPlaceIcon(FvwmWindow *t)
                   > dim[2].end_at)        /* check back edge */
                : (dim[2].start_at + dim[2].nom_dimension
                   < dim[2].end_at))       /* check front edge */
-              && (!loc_ok)) {             /* nothing found yet */
+	      && (!loc_ok)
+	      && (!loc_ok_wrong_screen2)) { /* nothing found yet */
           dim[2].real_start = dim[2].start_at; /* init */
           if (dim[2].start_at + dim[2].icon_dimension >
-              dim[2].screen_dimension - 2 + dim[2].base)
+              dim[2].screen_offset + dim[2].screen_dimension - 2 + dim[2].base)
 	  {
 	    /* if off screen, move on screen */
-            dim[2].real_start = dim[2].screen_dimension
+            dim[2].real_start = dim[2].screen_offset + dim[2].screen_dimension
               - dim[2].icon_dimension + dim[2].base;
           } /* end off screen */
-          if (dim[2].start_at < dim[2].base)
+          if (dim[2].start_at < dim[2].screen_offset + dim[2].base)
 	  {
 	    /* if off other edge, move on screen */
-	    dim[2].real_start = dim[2].base;
+	    dim[2].real_start = dim[2].screen_offset + dim[2].base;
           } /* end off other edge */
 
           if (HRZ_FILL)
@@ -763,10 +962,17 @@ void AutoPlaceIcon(FvwmWindow *t)
           }
 
 	  /* this may be a good location */
-          loc_ok = True;
+	  if (FScreenIsRectangleOnScreen(&fscr, FSCREEN_XYPOS, &ref))
+	  {
+	    loc_ok = True;
+	  }
+	  else
+	  {
+	    loc_ok_wrong_screen2 = True;
+	  }
           test_window = Scr.FvwmRoot.next;
           while((test_window != (FvwmWindow *)0)
-                &&(loc_ok == True))
+                &&(loc_ok == True || loc_ok_wrong_screen2))
 	  {
 	    /* test overlap */
             if(test_window->Desk == t->Desk)
@@ -787,11 +993,16 @@ void AutoPlaceIcon(FvwmWindow *t)
 		{
 		  /* don't accept this location */
                   loc_ok = False;
+                  loc_ok_wrong_screen2 = False;
                 } /* end if icons overlap */
               } /* end if its an icon */
             } /* end if same desk */
             test_window = test_window->next;
           } /* end while icons that may overlap */
+	  if (loc_ok_wrong_screen2)
+	  {
+	    loc_ok_wrong_screen = True;
+	  }
 	  /* Grid inner value & direction */
           dim[2].start_at += dim[2].step;
         } /* end while room inner dimension */
@@ -799,7 +1010,7 @@ void AutoPlaceIcon(FvwmWindow *t)
         dim[1].start_at += dim[1].step;
       } /* end while room outer dimension */
     } /* end for all icon boxes, or found space */
-    if(loc_ok == False)
+    if(!loc_ok && !loc_ok_wrong_screen)
       /* If icon never found a home just leave it */
       return;
     t->icon_g.x = real_x;
@@ -843,11 +1054,18 @@ static int
 do_all_iconboxes(FvwmWindow *t, icon_boxes **icon_boxes_ptr)
 {
   if (global_icon_box_ptr == 0) {       /* if first time */
-    /* Right now, the global box is hard-coded, fills the screen,
-       uses an 80x80 grid, and fills top-bottom, left-right */
+    int sx;
+    int sy;
+    int sw;
+    int sh;
+    /* Right now, the global box is hard-coded, fills the primary screen,
+     * uses an 80x80 grid, and fills top-bottom, left-right */
+    FScreenGetScrRect(NULL, FSCREEN_PRIMARY, &sx, &sy, &sw, &sh);
     global_icon_box_ptr = calloc(1, sizeof(icon_boxes));
-    global_icon_box_ptr->IconBox[2] = Scr.MyDisplayHeight;
-    global_icon_box_ptr->IconBox[3] = Scr.MyDisplayWidth;
+    global_icon_box_ptr->IconBox[0] = sx;
+    global_icon_box_ptr->IconBox[1] = sy;
+    global_icon_box_ptr->IconBox[2] = sx + sw;
+    global_icon_box_ptr->IconBox[3] = sy + sh;
     global_icon_box_ptr->IconGrid[0] = 80;
     global_icon_box_ptr->IconGrid[1] = 80;
     global_icon_box_ptr->IconFlags = ICONFILLHRZ;
@@ -909,9 +1127,9 @@ static void GetBitmapFile(FvwmWindow *tmp_win)
  * Looks for a color XPM icon file
  *
  ****************************************************************************/
+#ifdef XPM
 static void GetXPMFile(FvwmWindow *tmp_win)
 {
-#ifdef XPM
   XpmAttributes xpm_attributes;
   char *path = NULL;
   XpmImage my_image;
@@ -953,15 +1171,13 @@ static void GetXPMFile(FvwmWindow *tmp_win)
   SET_PIXMAP_OURS(tmp_win, 1);
   tmp_win->iconDepth = Pdepth;
 
-#ifdef SHAPE
-  if (ShapesSupported && tmp_win->icon_maskPixmap)
+  if (FShapesSupported && tmp_win->icon_maskPixmap)
     SET_ICON_SHAPED(tmp_win, 1);
-#endif
 
   XpmFreeXpmImage(&my_image);
 
-#endif /* XPM */
 }
+#endif /* XPM */
 
 /****************************************************************************
  *
@@ -988,8 +1204,7 @@ void GetIconWindow(FvwmWindow *tmp_win)
    * and define the cursor for it).
    */
   tmp_win->icon_pixmap_w = tmp_win->wmhints->icon_window;
-#ifdef SHAPE
-  if (ShapesSupported)
+  if (FShapesSupported)
   {
     if (tmp_win->wmhints->flags & IconMaskHint)
     {
@@ -997,7 +1212,6 @@ void GetIconWindow(FvwmWindow *tmp_win)
       tmp_win->icon_maskPixmap = tmp_win->wmhints->icon_mask;
     }
   }
-#endif
   /* Make sure that the window is a child of the root window ! */
   /* Olwais screws this up, maybe others do too! */
   XReparentWindow(dpy, tmp_win->icon_pixmap_w, Scr.Root, 0,0);
@@ -1033,8 +1247,7 @@ void GetIconBitmap(FvwmWindow *tmp_win)
   tmp_win->icon_p_height = height;
   tmp_win->iconDepth = depth;
 
-#ifdef SHAPE
-  if (ShapesSupported)
+  if (FShapesSupported)
   {
     if (tmp_win->wmhints->flags & IconMaskHint)
     {
@@ -1042,7 +1255,6 @@ void GetIconBitmap(FvwmWindow *tmp_win)
       tmp_win->icon_maskPixmap = tmp_win->wmhints->icon_mask;
     }
   }
-#endif
 
   SET_PIXMAP_OURS(tmp_win, 0);
 }
@@ -1058,17 +1270,16 @@ void GetIconBitmap(FvwmWindow *tmp_win)
 void DeIconify(FvwmWindow *tmp_win)
 {
   FvwmWindow *t,*tmp;
+  FvwmWindow *sf = get_focus_window();
 
-  if(!tmp_win)
+  if (!tmp_win)
     return;
-
   if (IS_MAP_PENDING(tmp_win))
   {
     /* final state: de-iconified */
     SET_ICONIFY_AFTER_MAP(tmp_win, 0);
     return;
   }
-
   while (IS_ICONIFIED_BY_PARENT(tmp_win))
   {
     for (t = Scr.FvwmRoot.next; t != NULL; t = t->next)
@@ -1079,12 +1290,19 @@ void DeIconify(FvwmWindow *tmp_win)
   }
 
   /* AS dje  RaiseWindow(tmp_win); */
+
+  mark_transient_subtree(tmp_win, MARK_ALL_LAYERS, MARK_ALL, False, True);
+  if (tmp_win == sf)
+  {
+    /* take away the focus before mapping */
+    DeleteFocus(0);
+  }
   /* now de-iconify transients */
   for (t = Scr.FvwmRoot.next; t != NULL; t = t->next)
   {
-    if ((t == tmp_win)||
-	((IS_TRANSIENT(t)) && (t->transientfor == tmp_win->w)))
+    if (t == tmp_win || IS_IN_TRANSIENT_SUBTREE(t))
     {
+      SET_IN_TRANSIENT_SUBTREE(t, 0);
       SET_MAPPED(t, 1);
       SET_ICONIFIED_BY_PARENT(t, 0);
       if(Scr.Hilite == t)
@@ -1098,7 +1316,7 @@ void DeIconify(FvwmWindow *tmp_win)
       XFlush(dpy);
       /* End AS */
       XMapWindow(dpy, t->w);
-      if(t->Desk == Scr.CurrentDesk)
+      if (t->Desk == Scr.CurrentDesk)
       {
 	rectangle r;
 
@@ -1106,9 +1324,10 @@ void DeIconify(FvwmWindow *tmp_win)
 	r.y = t->icon_g.y;
 	r.width = t->icon_g.width;
 	r.height = t->icon_p_height+t->icon_g.height;
-	/* update frame geometry in case the icon was moved over a page
-	 * boundary*/
-	update_relative_geometry(t);
+	/* update absoluthe geometry in case the icon was moved over a page
+	 * boundary; the move code already takes care of keeping the frame
+	 * geometry up to date */
+	update_absolute_geometry(t);
 	if (IsRectangleOnThisPage(&r, t->Desk) &&
 	    !IsRectangleOnThisPage(&(t->frame_g), t->Desk))
 	{
@@ -1148,7 +1367,7 @@ void DeIconify(FvwmWindow *tmp_win)
 			(unsigned long)t,
 			t->icon_g.x, t->icon_g.y,
 			t->icon_p_width,
-			t->icon_p_height+t->icon_g.height);
+			t->icon_p_height + t->icon_g.height);
       }
       if(t->Desk == Scr.CurrentDesk)
       {
@@ -1172,9 +1391,26 @@ void DeIconify(FvwmWindow *tmp_win)
 #if 1
   RaiseWindow(tmp_win); /* moved dje */
 #endif
-
-  if(HAS_CLICK_FOCUS(tmp_win))
+  if (sf)
+  {
+    focus_grab_buttons(sf, True);
+  }
+  if (sf == tmp_win)
+  {
+    /* update the focus to make sure the application knows its state */
+    if (HAS_CLICK_FOCUS(tmp_win) || HAS_SLOPPY_FOCUS(tmp_win))
+    {
+      SetFocusWindow(tmp_win, 0);
+    }
+  }
+  else if (HAS_CLICK_FOCUS(tmp_win))
+  {
+#if 0
     FocusOn(tmp_win, TRUE, "");
+#else
+    SetFocusWindow(tmp_win, 1);
+#endif
+  }
   GNOME_SetWinArea(tmp_win);
 
   return;
@@ -1189,35 +1425,42 @@ void DeIconify(FvwmWindow *tmp_win)
 void Iconify(FvwmWindow *tmp_win, int def_x, int def_y)
 {
   FvwmWindow *t;
+  FvwmWindow *sf;
   XWindowAttributes winattrs = {0};
   unsigned long eventMask;
 
   if(!tmp_win)
     return;
-
   if (IS_MAP_PENDING(tmp_win))
   {
     /* final state: iconified */
     SET_ICONIFY_AFTER_MAP(tmp_win, 1);
     return;
   }
-
   if (!XGetWindowAttributes(dpy, tmp_win->w, &winattrs))
   {
     return;
   }
   eventMask = winattrs.your_event_mask;
-
-  if((tmp_win == Scr.Hilite)&&(HAS_CLICK_FOCUS(tmp_win))&&(tmp_win->next))
+#if 0
+  if (tmp_win == Scr.Hilite && HAS_CLICK_FOCUS(tmp_win) && tmp_win->next)
   {
     SetFocusWindow(tmp_win->next, 1);
   }
+#endif
 
+  mark_transient_subtree(tmp_win, MARK_ALL_LAYERS, MARK_ALL, False, True);
+  sf = get_focus_window();
+  if (sf && IS_IN_TRANSIENT_SUBTREE(sf))
+  {
+    restore_focus_after_unmap(sf, True);
+  }
   /* iconify transients first */
   for (t = Scr.FvwmRoot.next; t != NULL; t = t->next)
   {
-    if ((t==tmp_win)||((IS_TRANSIENT(t)) && (t->transientfor == tmp_win->w)))
+    if (t == tmp_win || IS_IN_TRANSIENT_SUBTREE(t))
     {
+      SET_IN_TRANSIENT_SUBTREE(t, 0);
       SET_ICON_ENTERED(t, 0);
       /*
        * Prevent the receipt of an UnmapNotify, since that would
@@ -1236,7 +1479,7 @@ void Iconify(FvwmWindow *tmp_win, int def_x, int def_y)
 
       SetMapStateProp(t, IconicState);
       DrawDecorations(t, DRAW_ALL, False, False, None);
-      if (t == tmp_win)
+      if (t == tmp_win && !IS_ICONIFIED_BY_PARENT(tmp_win))
       {
 	SET_DEICONIFY_PENDING(t, 1);
       }
@@ -1261,16 +1504,17 @@ void Iconify(FvwmWindow *tmp_win, int def_x, int def_y)
   if (IS_ICONIFIED_BY_PARENT(tmp_win))
     return;
 
-  if (tmp_win->icon_w == None)
+  if (tmp_win->icon_w == None || HAS_ICON_CHANGED(tmp_win))
   {
     if(IS_ICON_MOVED(tmp_win))
       CreateIconWindow(tmp_win,tmp_win->icon_g.x,tmp_win->icon_g.y);
     else
       CreateIconWindow(tmp_win, def_x, def_y);
+    SET_HAS_ICON_CHANGED(tmp_win, 0);
   }
 
   /* if no pixmap we want icon width to change to text width every iconify */
-  if( (tmp_win->icon_w != None) && (tmp_win->icon_pixmap_w == None) )
+  if (tmp_win->icon_w && !tmp_win->icon_pixmap_w)
   {
     if (HAS_NO_ICON_TITLE(tmp_win))
     {
@@ -1322,8 +1566,11 @@ void Iconify(FvwmWindow *tmp_win, int def_x, int def_y)
   {
     LowerWindow(tmp_win);
   }
-
-  if(tmp_win->Desk == Scr.CurrentDesk)
+  if (IS_ICON_STICKY(tmp_win) || IS_STICKY(tmp_win))
+  {
+    tmp_win->Desk = Scr.CurrentDesk;
+  }
+  if (tmp_win->Desk == Scr.CurrentDesk)
   {
     if (tmp_win->icon_w != None)
       XMapWindow(dpy, tmp_win->icon_w);
@@ -1331,18 +1578,28 @@ void Iconify(FvwmWindow *tmp_win, int def_x, int def_y)
     if(tmp_win->icon_pixmap_w != None)
       XMapWindow(dpy, tmp_win->icon_pixmap_w);
   }
-  if(HAS_CLICK_FOCUS(tmp_win) || HAS_SLOPPY_FOCUS(tmp_win))
+#if 0
+  if (HAS_CLICK_FOCUS(tmp_win) || HAS_SLOPPY_FOCUS(tmp_win))
   {
-    if (tmp_win == Scr.Focus)
+    if (tmp_win == get_focus_window())
     {
-      if(Scr.PreviousFocus == Scr.Focus)
-	Scr.PreviousFocus = NULL;
-      if(HAS_CLICK_FOCUS(tmp_win) && tmp_win->next)
+      if (HAS_CLICK_FOCUS(tmp_win) && tmp_win->next)
+      {
 	SetFocusWindow(tmp_win->next, 1);
+      }
       else
+      {
 	DeleteFocus(1);
+      }
     }
   }
+#endif
+  if ((sf = get_focus_window()))
+  {
+    focus_grab_buttons(sf, True);
+  }
+
+  return;
 }
 
 
@@ -1368,9 +1625,11 @@ void SetMapStateProp(FvwmWindow *tmp_win, int state)
 }
 
 
-void iconify_function(F_CMD_ARGS)
+void CMD_Iconify(F_CMD_ARGS)
 {
   int toggle;
+  int x;
+  int y;
 
   if (DeferExecution(eventp,&w,&tmp_win,&context, CRS_SELECT, ButtonRelease))
     return;
@@ -1405,7 +1664,10 @@ void iconify_function(F_CMD_ARGS)
 	XBell(dpy, 0);
 	return;
       }
-      Iconify(tmp_win,eventp->xbutton.x_root-5,eventp->xbutton.y_root-5);
+      x = 0;
+      y = 0;
+      GetLocationFromEventOrQuery(dpy, Scr.Root, eventp, &x, &y);
+      Iconify(tmp_win, x, y);
     }
   }
 }
