@@ -24,6 +24,8 @@
 #include <X11/Intrinsic.h>              /* needed for xpm.h and Pixel defn */
 #include <ctype.h>
 
+#include "fvwmrect.h"
+
 /* Allow GCC extensions to work, if you have GCC */
 
 #ifndef __attribute__
@@ -75,12 +77,14 @@
                       f_db_print _x; }while(0)
 # define DB(_x)     DBL(2,_x)
 
-struct f_db_info {
+struct f_db_info
+{
   const char *filenm;
   const char *funcnm;
   const char *module;
   long lineno;
 };
+
 extern struct f_db_info f_db_info;
 extern int f_db_level;
 
@@ -89,6 +93,12 @@ extern void f_db_print(const char *fmt, ...)
 #endif
 
 
+
+/***********************************************************************
+ * typedefs
+ ***********************************************************************/
+
+
 /***********************************************************************
  * Replacements for missing system calls.
  ***********************************************************************/
@@ -244,6 +254,10 @@ void PrintXErrorAndCoredump(Display *dpy, XErrorEvent *error, char *MyName);
  */
 Bool GetLocationFromEventOrQuery(Display *dpy, Window w, XEvent *eventp,
 				 int *ret_x, int *ret_y);
+/*
+ * Return the subwindow member of an event if the event type has one.
+ */
+Window GetSubwindowFromEvent(Display *dpy, XEvent *eventp);
 
 /***********************************************************************
  * Wrappers around Xrm routines (XResources.c)
@@ -254,8 +268,8 @@ void MergeCmdLineResources(XrmDatabase *pdb, XrmOptionDescList opts,
 			   Bool fNoDefaults);
 Bool MergeConfigLineResource(XrmDatabase *pdb, char *line, char *prefix,
 			     char *bindstr);
-Bool GetResourceString(XrmDatabase db, const char *resource,
-		       const char *prefix, char **val);
+Bool GetResourceString(
+  XrmDatabase db, const char *resource, const char *prefix, XrmValue *xval);
 
 
 /***********************************************************************
@@ -264,6 +278,8 @@ Bool GetResourceString(XrmDatabase db, const char *resource,
 
 void RelieveRectangle(Display *dpy, Drawable d, int x,int y,int w,int h,
 		      GC ReliefGC, GC ShadowGC, int line_width);
+void RelieveRectangle2(Display *dpy, Drawable d, int x,int y,int w,int h,
+		      GC ReliefGC, GC ShadowGC, int line_width);
 
 Pixmap CreateStretchXPixmap(Display *dpy, Pixmap src, int src_width,
 			    int src_height, int src_depth,
@@ -318,13 +334,10 @@ Bool CalculateGradientDimensions(Display *dpy, Drawable d, int ncolors,
 				 char type, unsigned int *width_ret,
 				 unsigned int *height_ret);
 
-Drawable CreateGradientPixmap(Display *dpy, Drawable d, GC gc,
-			      int type, int g_width, int g_height,
-			      int ncolors, Pixel *pixels,
-			      Drawable in_drawable,
-			      int d_x, int d_y,
-			      int d_width,
-			      int d_height);
+Drawable CreateGradientPixmap(
+  Display *dpy, Drawable d, GC gc, int type, int g_width, int g_height,
+  int ncolors, Pixel *pixels, Drawable in_drawable, int d_x, int d_y,
+  int d_width, int d_height, XRectangle *rclip);
 
 Pixmap CreateGradientPixmapFromString(Display *dpy, Drawable d, GC gc,
 				      int type, char *action,
@@ -409,11 +422,16 @@ typedef struct Binding
 
 Bool ParseContext(char *in_context, int *out_context_mask);
 Bool ParseModifiers(char *in_modifiers, int *out_modifier_mask);
-int AddBinding(Display *dpy, Binding **pblist, BindingType type,
-	       STROKE_ARG(void *stroke)
-	       int button, KeySym keysym, char *key_name,
-	       int modifiers, int contexts, void *action, void *action2);
-void FreeBinding(Binding *b);
+void CollectBindingList(
+  Display *dpy, Binding **pblist_src, Binding **pblist_dest, BindingType type,
+  STROKE_ARG(void *stroke)
+  int button, KeySym keysym, int modifiers, int contexts);
+int AddBinding(
+  Display *dpy, Binding **pblist, BindingType type, STROKE_ARG(void *stroke)
+  int button, KeySym keysym, char *key_name, int modifiers, int contexts,
+  void *action, void *action2);
+void FreeBindingStruct(Binding *b);
+void FreeBindingList(Binding *b);
 void RemoveBinding(Binding **pblist, Binding *b, Binding *prev);
 Bool RemoveMatchingBinding(
   Display *dpy, Binding **pblist, BindingType type, STROKE_ARG(char *stroke)
@@ -429,7 +447,7 @@ Bool MatchBinding(Display *dpy, Binding *b,
 		  unsigned int modifier,unsigned int dead_modifiers,
 		  int Context, BindingType type);
 Bool MatchBindingExactly(
-  Display *dpy, Binding *b, STROKE_ARG(void *stroke)
+  Binding *b, STROKE_ARG(void *stroke)
   int button, KeyCode keycode, unsigned int modifier, int Context,
   BindingType type);
 void GrabWindowKey(Display *dpy, Window w, Binding *binding,
@@ -448,6 +466,9 @@ void GrabAllWindowKeysAndButtons(Display *dpy, Window w, Binding *blist,
 				 unsigned int contexts,
 				 unsigned int dead_modifiers,
 				 Cursor cursor, Bool fGrab);
+void GrabWindowKeyOrButton(
+  Display *dpy, Window w, Binding *binding, unsigned int contexts,
+  unsigned int dead_modifiers, Cursor cursor, Bool fGrab);
 KeySym FvwmStringToKeysym(Display *dpy, char *key);
 
 /***********************************************************************
@@ -464,6 +485,11 @@ void fvwmlib_get_target_window(
 
 Window fvwmlib_client_window(Display *dpy, Window input);
 
+/***********************************************************************
+ * Cursor.c
+ ***********************************************************************/
+
+int fvwmCursorNameToIndex (char *cursor_name);
 
 /* Set up heap debugging library dmalloc.  */
 
